@@ -11,14 +11,17 @@ import sys
 import tempfile
 import time
 import unittest
+from collections import defaultdict
+from random import shuffle
 
 sys.path.append(os.path.dirname(os.path.dirname(__file__)))
 
-from tests.utils import tester
+from tests.utils import tester, subscript_item, in_finally
 
 PYPY = 'pypy' in sys.version.lower()
 
-from executing import Source, only, PY3, NotOneValueFound, get_instructions
+from executing import Source, only, NotOneValueFound
+from executing.executing import PY3, get_instructions
 
 
 class TestStuff(unittest.TestCase):
@@ -297,17 +300,35 @@ class TestStuff(unittest.TestCase):
             self.assertTrue(isinstance(ex.node, ast.BinOp))
             self.assertEqual(ex.text(), "134895 / 0")
 
+    def test_retry_cache(self):
+        _, filename = tempfile.mkstemp()
+
+        def check(x):
+            source = 'tester(6)\n%s\ntester(9)' % list(range(x))
+            code = compile(source, filename, 'exec')
+            with open(filename, 'w') as outfile:
+                outfile.write(source)
+            exec(code, globals(), locals())
+
+        check(3)
+        check(5)
+
 
 def is_unary_not(node):
     return isinstance(node, ast.UnaryOp) and isinstance(node.op, ast.Not)
 
 
+class TimeOut(Exception):
+    pass
+
+
 @unittest.skipUnless(
     os.getenv('EXECUTING_SLOW_TESTS'),
     'These tests are very slow, enable them explicitly',
 )
 class TestFiles(unittest.TestCase):
     def test_files(self):
+        self.start_time = time.time()
         root_dir = os.path.dirname(__file__)
         samples_dir = os.path.join(root_dir, 'samples')
         result_filename = PYPY * 'pypy' + sys.version[:3] + '.json'
@@ -325,7 +346,9 @@ class TestFiles(unittest.TestCase):
             with open(result_filename, 'r') as infile:
                 self.assertEqual(result, json.load(infile))
 
-        for module in list(sys.modules.values()):
+        modules = list(sys.modules.values())
+        shuffle(modules)
+        for module in modules:
             try:
                 filename = inspect.getsourcefile(module)
             except TypeError:
@@ -336,15 +359,23 @@ class TestFiles(unittest.TestCase):
 
             filename = os.path.abspath(filename)
 
-            if 'executing' in filename:
+            if (
+                    # The sentinel actually appearing in code messes things up
+                    'executing' in filename
+                    # A file that's particularly slow
+                    or 'errorcodes.py' in filename
+            ):
                 continue
 
-            self.check_filename(filename)
+            try:
+                self.check_filename(filename)
+            except TimeOut:
+                print("Time's up")
 
     def check_filename(self, filename):
         print(filename)
         source = Source.for_filename(filename)
-        nodes = {}
+        nodes = defaultdict(list)
         for node in ast.walk(source.tree):
             if isinstance(node, (
                     ast.UnaryOp,
@@ -354,23 +385,23 @@ class TestFiles(unittest.TestCase):
                     ast.Compare,
                     ast.Attribute
             )):
-                nodes[node] = None
+                nodes[node] = []
 
         code = compile(source.tree, source.filename, 'exec')
         result = list(self.check_code(code, nodes))
 
         if not re.search(r'^\s*if 0(:| and )', source.text, re.MULTILINE):
-            for node, value in nodes.items():
+            for node, values in nodes.items():
                 if is_unary_not(node):
                     continue
 
                 if isinstance(getattr(node, 'ctx', None), (ast.Store, ast.Del)):
-                    assert value is None
+                    assert not values
                     continue
 
                 if isinstance(node, ast.Compare):
                     if len(node.ops) > 1:
-                        assert value is None
+                        assert not values
                         continue
 
                     if is_unary_not(node.parent) and isinstance(node.ops[0], (ast.In, ast.Is)):
@@ -379,8 +410,14 @@ class TestFiles(unittest.TestCase):
                 if is_literal(node):
                     continue
 
-                if value is None:
-                    print(source.text, '---', node_string(source, node), file=sys.stderr, sep='\n')
+                if sys.version_info >= (3, 9) and in_finally(node):
+                    correct = len(values) > 1
+                else:
+                    correct = len(values) == 1
+
+                if not correct:
+                    print(source.text, '---', node_string(source, node), node.lineno,
+                          len(values), correct, values, file=sys.stderr, sep='\n')
                     self.fail()
 
         return result
@@ -390,10 +427,14 @@ class TestFiles(unittest.TestCase):
         instructions = get_instructions(code)
         lineno = None
         for inst in instructions:
+            if time.time() - self.start_time > 45 * 60:
+                # Avoid travis time limit of 50 minutes
+                raise TimeOut
+
             lineno = linestarts.get(inst.offset, lineno)
             if not inst.opname.startswith((
                     'BINARY_', 'UNARY_', 'LOAD_ATTR', 'LOAD_METHOD', 'LOOKUP_METHOD',
-                    'SLICE+', 'COMPARE_OP', 'CALL_',
+                    'SLICE+', 'COMPARE_OP', 'CALL_', 'IS_OP', 'CONTAINS_OP',
             )):
                 continue
             frame = C()
@@ -413,16 +454,11 @@ class TestFiles(unittest.TestCase):
                     if isinstance(only(source.statements_at_line(lineno)), (ast.AugAssign, ast.Import)):
                         continue
                     raise
-
-                try:
-                    self.assertIsNone(nodes[node])
-                except KeyError:
-                    print(ast.dump(source.tree), list(ast.walk(source.tree)), nodes, node, ast.dump(node), file=sys.stderr, sep='\n')
             except Exception:
                 print(source.text, lineno, inst, node and ast.dump(node), code, file=sys.stderr, sep='\n')
                 raise
 
-            nodes[node] = (inst, frame.__dict__)
+            nodes[node].append((inst, frame.__dict__))
 
             yield [inst.opname, node_string(source, node)]
 
@@ -447,10 +483,7 @@ def is_literal(node):
         return all(map(is_literal, [node.left] + node.comparators))
 
     if isinstance(node, ast.Subscript) and is_literal(node.value):
-        if isinstance(node.slice, ast.Index):
-            return is_literal(node.slice.value)
-        else:
-            assert isinstance(node.slice, ast.Slice)
+        if isinstance(node.slice, ast.Slice):
             return all(
                 x is None or is_literal(x)
                 for x in [
@@ -459,6 +492,8 @@ def is_literal(node):
                     node.slice.step,
                 ]
             )
+        else:
+            return is_literal(subscript_item(node))
 
     try:
         ast.literal_eval(node)
@@ -545,3 +580,4 @@ def decorator_with_args(*_, **__):
 
 if __name__ == '__main__':
     unittest.main()
+
